import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'dart:ui' as ui;

/// The breakpoints are
/// 0 logical pixel<= smallest width < 480 logical pixel => MOBILE_PORTRAIT
/// 480 logical pixel<= smallest width < 720 logical pixel => TABLET_PORTRAIT
/// 720 logical pixel<= smallest width => WEB = (mobile landscape) = (tablet landscape)
enum DEVICE {
  MOBILE_PORTRAIT,
  TABLET_PORTRAIT,
  WEB, // web is equal to mobile landscape and tablet landscape
}

/// Class use for beautiful responsive design both on tablets and mobiles
/// This class specially used for developing an app for all devices without
/// having specific different designs
abstract class ResponsiveSizeUtils {
  /// design draft device width
  static late double _designedDeviceWidth;

  /// design draft device height
  static late double _designedDeviceHeight;

  /// device width
  static late double _screenWidth;

  /// device height
  static late double _screenHeight;

  /// device orientation
  static late Orientation _orientation;

  /// device status bar
  static late double _statusBarHeight;

  /// device system bar height
  static late double _systemBottomNavigationBarHeight;

  // device status bar height
  static double get statusBarHeight => _statusBarHeight;

  /// device system bar height
  static double get systemBottomNavigationBarHeight =>
      _systemBottomNavigationBarHeight;

  /// assign device values
  static void _calculate({
    required double screenWidth,
    required double screenHeight,
    required Orientation orientation,
  }) {
    _screenWidth = screenWidth;
    _screenHeight = screenHeight;
    _orientation = orientation;
  }

  /// assign design values
  static void _initDesignedSize({
    required double designedDeviceWidth,
    required double designedDeviceHeight,
  }) {
    _designedDeviceWidth = designedDeviceWidth;
    _designedDeviceHeight = designedDeviceHeight;
  }

  /// return a reasonable width `value`.
  ///
  /// if `maxWidth` is null then the `value` will be at most `designedWidth`
  /// logical pixel otherwise, the `value` will be at most `maxWidth` logical pixels.
  ///
  /// `value` can be any double between zero and maximum width.
  ///
  static double width(
    double designedWidth, {
    double? maxWidth,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _width = _screenWidth * designedWidth / _designedDeviceWidth;
      if (maxWidth == null) {
        if (_width <= designedWidth) {
          return _width;
        } else {
          return designedWidth;
        }
      } else {
        return _comparator(
          value: designedWidth,
          maxValue: maxWidth,
        );
      }
    } else {
      final _width = _screenWidth * designedWidth / _designedDeviceHeight;
      if (maxWidth == null) {
        if (_width <= designedWidth) {
          return _width;
        } else {
          return designedWidth;
        }
      } else {
        return _comparator(
          value: designedWidth,
          maxValue: maxWidth,
        );
      }
    }
  }

  /// return a reasonable height `value`.
  ///
  /// if `maxHeight` is null then the `value` will be at most `designedHeight`
  /// logical pixel otherwise, the `value` will be at most `maxHeight` logical pixels.
  ///
  /// `value` can be any double between zero and maximum height.
  ///
  static double height(
    double designedHeight, {
    double? maxHeight,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _height = _screenHeight * designedHeight / _designedDeviceHeight;
      if (maxHeight == null) {
        if (_height <= designedHeight) {
          return _height;
        } else {
          return designedHeight;
        }
      } else {
        return _comparator(
          value: _height,
          maxValue: maxHeight,
        );
      }
    } else {
      final _height = _screenHeight * designedHeight / _designedDeviceWidth;
      if (maxHeight == null) {
        if (_height <= designedHeight) {
          return _height;
        } else {
          return designedHeight;
        }
      } else {
        return _comparator(
          value: _height,
          maxValue: maxHeight,
        );
      }
    }
  }

  /// return `width` of screen
  static double fullWidth() {
    return _screenWidth;
  }

  /// return `height` of screen
  static double fullHeight() {
    return _screenHeight;
  }

  /// With this class you can have
  static double? widthSpacing({
    required double spaceSizeBeforeBreakPoint,
    required double spaceSizeAfterBreakPoint,
  }) {
    final _screenWidth = ResponsiveSizeUtils._screenWidth;
    return _screenWidth <= _designedDeviceWidth
        ? width(
            spaceSizeBeforeBreakPoint,
          )
        : width(
            spaceSizeAfterBreakPoint,
          );
  }

  static double heightSpacing({
    required double spaceSizeBeforeBreakPoint,
    required double spaceSizeAfterBreakPoint,
  }) {
    return _screenHeight <= _designedDeviceHeight
        ? height(
            spaceSizeBeforeBreakPoint,
          )
        : height(
            spaceSizeAfterBreakPoint,
          );
  }

  /// For calculating how much
  static double widthFraction({
    required double fraction,
  }) {
    assert(
      fraction < 1,
      'fraction must be lower than 1',
    );
    return _screenWidth * fraction;
  }

  /// Return the
  static double heightFraction({
    required double fraction,
  }) {
    assert(
      fraction < 1,
      'fraction must be lower than 1',
    );
    return _screenHeight * fraction;
  }

  /// Show if device is in portrait mode
  static bool isPortrait() {
    return _orientation == Orientation.portrait;
  }

  /// Return smallest width of device
  static double smallestWidth() {
    final _width = _screenWidth;
    final _height = _screenHeight;
    return _width <= _height ? _width : _height;
  }

  /// Return [DEVICE] type
  /// 0 <= [Mobile portrait] <= 480 <= [Tablet portrait] <= 720 <= [Web]
  /// [Web] is also equal to [mobile landscape] and [tablet landscape] and also [desktop]
  static DEVICE device() {
    final _smallestWidth = smallestWidth();
    if (_smallestWidth <= 480.0) {
      return DEVICE.MOBILE_PORTRAIT;
    } else if (_smallestWidth <= 720.0) {
      return DEVICE.TABLET_PORTRAIT;
    } else {
      return DEVICE.WEB;
    }
  }

  /// Compare two value and return the smallest
  static double _comparator({
    required double value,
    required double maxValue,
  }) {
    return value <= maxValue ? value : maxValue;
  }

  /// With this widget you can have three different layouts for each kind of
  /// device
  static Widget responsiveWidgetAccordingToDevice({
    required Widget mobilePortrait,
    required Widget tabletPortrait,
    required Widget web,
  }) {
    switch (device()) {
      case DEVICE.WEB:
        return web;
      case DEVICE.TABLET_PORTRAIT:
        return tabletPortrait;
      default:
        return mobilePortrait;
    }
  }

  /// This class gives you a reasonable font Size
  /// Use this class with fontSize and fontHeight property of TextStyle
  static double responsiveFont({
    required double fontSize,
    double? maxFontSize,
  }) {
    final orientation = _orientation;
    if (orientation == Orientation.portrait) {
      final _calculatedFontSize =
          _screenHeight * fontSize / _designedDeviceHeight;
      if (maxFontSize == null) {
        if (_calculatedFontSize <= fontSize) {
          return _calculatedFontSize;
        } else {
          return fontSize;
        }
      } else {
        return _comparator(
          value: _calculatedFontSize,
          maxValue: maxFontSize,
        );
      }
    } else {
      final _calculatedFontSize =
          _screenHeight * fontSize / _designedDeviceWidth;
      if (maxFontSize == null) {
        if (_calculatedFontSize <= fontSize) {
          return _calculatedFontSize;
        } else {
          return fontSize;
        }
      } else {
        return _comparator(
          value: _calculatedFontSize,
          maxValue: maxFontSize,
        );
      }
    }
  }
}

class ResponsiveWidget extends StatelessWidget {
  /// A helper class that init the [ResponsiveSizeUtils] class
  /// wrap build method of first widget
  ResponsiveWidget({
    Key? key,
    required this.builder,
    required this.designedSize,
    this.init,
  }) : super(key: key) {
    ResponsiveSizeUtils._initDesignedSize(
      designedDeviceWidth: designedSize.width,
      designedDeviceHeight: designedSize.height,
    );
  }

  /// builder function for rebuilding the rest of UI
  final Widget Function() builder;

  /// The [Size] of device in designs drafts.
  final Size designedSize;

  /// This function call after [ResponsiveSizeUtils] initialization and before
  /// [builder] function. Use this for assigning font sizes or any other related
  /// sizing before whole app build.
  final void Function()? init;

  @override
  Widget build(
    BuildContext context,
  ) {
    return LayoutBuilder(
      builder: (
        context,
        constraints,
      ) {
        final Orientation orientation =
            constraints.maxWidth > constraints.maxHeight
                ? Orientation.landscape
                : Orientation.portrait;
        final window = WidgetsBinding.instance?.window ?? ui.window;
        ResponsiveSizeUtils._statusBarHeight = window.padding.top;
        ResponsiveSizeUtils._systemBottomNavigationBarHeight =
            window.padding.bottom;
        if (orientation == Orientation.portrait) {
          ResponsiveSizeUtils._calculate(
            screenWidth: constraints.maxWidth,
            screenHeight: constraints.maxHeight,
            orientation: orientation,
          );
        } else {
          if (kIsWeb) {
            ResponsiveSizeUtils._calculate(
              screenWidth: constraints.maxWidth,
              screenHeight: constraints.maxHeight,
              orientation: orientation,
            );
          } else {
            ResponsiveSizeUtils._calculate(
              screenWidth: constraints.maxHeight,
              screenHeight: constraints.maxWidth,
              orientation: orientation,
            );
          }
        }
        if (init != null) {
          init!.call();
        }
        return builder();
      },
    );
  }
}
